__all__ = ('CommandParameter', 'get_command_parameters_for',)

from scarletio import CallableAnalyzer, RichAttributeErrorBaseType


TYPE_IDENTIFIER_STR = 1
TYPE_IDENTIFIER_INT = 2
TYPE_IDENTIFIER_FLOAT = 3
TYPE_IDENTIFIER_BOOL = 4

TYPE_TYPE_STR = str
TYPE_TYPE_INT = int
TYPE_TYPE_FLOAT = float
TYPE_TYPE_BOOL = bool

TYPE_NAME_STR = 'str'
TYPE_NAME_INT = 'int'
TYPE_NAME_FLOAT = 'float'
TYPE_NAME_BOOL = 'bool'

TYPE_TYPE_TO_IDENTIFIER = {
    TYPE_TYPE_STR: TYPE_IDENTIFIER_STR,
    TYPE_TYPE_INT: TYPE_IDENTIFIER_INT,
    TYPE_TYPE_FLOAT: TYPE_IDENTIFIER_FLOAT,
    TYPE_TYPE_BOOL: TYPE_IDENTIFIER_BOOL
}

TYPE_NAME_TO_IDENTIFIER = {
    TYPE_NAME_STR: TYPE_IDENTIFIER_STR,
    TYPE_NAME_INT: TYPE_IDENTIFIER_INT,
    TYPE_NAME_FLOAT: TYPE_IDENTIFIER_FLOAT,
    TYPE_NAME_BOOL: TYPE_IDENTIFIER_BOOL,
}


# revert the relation
TYPE_IDENTIFIER_TO_NAME = {identifier: name for name, identifier in TYPE_NAME_TO_IDENTIFIER.items()}


def parameter_value_converter_str(value):
    """
    Converts the given value to string.
    
    Parameters
    ----------
    value : `str`
        The value to convert to string.
    
    Returns
    -------
    value : `None`, `str`
        Returns `None` if conversion failed.
    """
    return value


def parameter_value_converter_int(value):
    """
    Converts the given value to int.
    
    Parameters
    ----------
    value : `str`
        The value to convert to string.
    
    Returns
    -------
    value : `None`, `int`
        Returns `None` if conversion failed.
    """
    try:
        value = int(value)
    except ValueError:
        value = None
    
    return value


def parameter_value_converter_float(value):
    """
    Converts the given value to float.
    
    Parameters
    ----------
    value : `str`
        The value to convert to string.
    
    Returns
    -------
    value : `None`, `float`
        Returns `None` if conversion failed.
    """
    try:
        value = float(value)
    except ValueError:
        value = None
    
    return value


def parameter_value_converter_bool(value):
    """
    Converts the given value to bool.
    
    Parameters
    ----------
    value : `bool`, `str`
        The value to convert to string.
    
    Returns
    -------
    value : `None`, `bool`
        Returns `None` if conversion failed.
    """
    if isinstance(value, bool):
        return value
    
    if isinstance(value, str):
        value = value.lower()
        if value == 'true':
            return True
        
        if value == 'false':
            return False
        
        return None
    
    return None


TYPE_IDENTIFIER_TO_CONVERTER = {
    TYPE_IDENTIFIER_STR: parameter_value_converter_str,
    TYPE_IDENTIFIER_INT: parameter_value_converter_int,
    TYPE_IDENTIFIER_FLOAT: parameter_value_converter_float,
    TYPE_IDENTIFIER_BOOL: parameter_value_converter_bool,
}


class CommandParameter(RichAttributeErrorBaseType):
    """
    Represents a command line command's parameter.
    
    Attributes
    ----------
    display_name : `str`
        The display name of the parameter.
    expected_type_identifier : `int`
        Internal identifier for the expected type accepted by the command.
    parameter : ``Parameter``
        The command function's parameter.
    """
    __slots__ = ('display_name', 'expected_type_identifier', 'parameter',)
    
    def __new__(cls, parameter):
        """
        Creates a new command line parameter.
        
        Parameters
        ----------
        parameter : ``Parameter``
            The respective function's parameter.
        
        Raises
        ------
        LookupError
            Could not match the parameter's annotation.
        TypeError
            Unhandled annotation type.
        ValueError
            Not allowed parameter specification.
        """
        if parameter.has_annotation:
            annotation = parameter.annotation
            
            if isinstance(annotation, str):
                try:
                    expected_type_identifier = TYPE_NAME_TO_IDENTIFIER[annotation]
                except KeyError:
                    raise LookupError(
                        f'Annotation has no converter. Got: {annotation!r}.'
                    ) from None
                    
            elif isinstance(annotation, type):
                try:
                    expected_type_identifier = TYPE_TYPE_TO_IDENTIFIER[annotation]
                except KeyError:
                    raise LookupError(
                        f'Annotation has no converter. Got: {annotation!r}.'
                    ) from None
            
            else:
                raise TypeError(
                    f'Annotation of unexpected type Got {annotation!r}.'
                )
        
        else:
            expected_type_identifier = TYPE_IDENTIFIER_STR
        
        if parameter.is_positional() and expected_type_identifier == TYPE_IDENTIFIER_BOOL:
            raise ValueError(
                f'Not allowed parameter specification. Positional parameters cannot be annotated as bool.'
                f'Got: parameter={parameter!r}'
            )
        
        display_name = '-'.join(parameter.name.replace('_', ' ').split())
        
        self = object.__new__(cls)
        self.display_name = display_name
        self.expected_type_identifier = expected_type_identifier
        self.parameter = parameter
        return self
    
    
    def __repr__(self):
        """Returns the command line parameter's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' ')
        repr_parts.append(self.display_name)
        
        repr_parts.append(', parameter = ')
        repr_parts.append(repr(self.parameter))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def get_converter(self):
        """
        Gets converter of the specified parameter type.
        
        Returns
        -------
        converter : `FunctionType`
        """
        return TYPE_IDENTIFIER_TO_CONVERTER[self.expected_type_identifier]
    
    
    def is_modifier(self):
        """
        Returns whether the parameter is a modifier parameter.
        
        Returns
        -------
        is_modifier : `bool`
        """
        return self.expected_type_identifier == TYPE_IDENTIFIER_BOOL
    
    
    def is_required(self):
        """
        Returns whether the parameter is required.
        """
        parameter = self.parameter
        if parameter.has_default:
            return False
        
        if parameter.is_args() or parameter.is_kwargs():
            return False
        
        return True


def get_command_parameters_for(function):
    """
    Gets the command parameter
    """
    analyzer = CallableAnalyzer(function)
    
    command_line_parameters = []
    
    for parameter in analyzer.iter_non_reserved_parameters():
        command_line_parameter = CommandParameter(parameter)
        command_line_parameters.append(command_line_parameter)
    
    return command_line_parameters



__all__ = ('AutoModerationActionType',)

from ...bases import Preinstance as P, PreinstancedBase

from ..action_metadata import (
    AutoModerationActionMetadataBase, AutoModerationActionMetadataBlock, AutoModerationActionMetadataSendAlertMessage,
    AutoModerationActionMetadataTimeout
)


class AutoModerationActionType(PreinstancedBase, value_type = int):
    """
    Represents an ``AutoModerationAction``'s type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the auto moderation action type.
    
    name : `str`
        The default name of the auto moderation action type.
    
    metadata_type : `type<<AutoModerationActionMetadataBase>`
        The action type's respective metadata type.
    
    Type Attributes
    ---------------
    Every predefined auto moderation action type is also stored as a type attribute:
    
    +---------------------------+---------------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | Type attribute name       | Name                      | Value     | Metadata type                                     | Description                                                           |
    +===========================+===========================+===========+===================================================+=======================================================================+
    | none                      | none                      | 0         | ``AutoModerationActionMetadataBase``              | N/A                                                                   |
    +---------------------------+---------------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | block_message             | block message             | 1         | ``AutoModerationActionMetadataBlock``             | Blocks the message's content according to the rule.                   |
    +---------------------------+---------------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | send_alert_message        | send alert message        | 2         | ``AutoModerationActionMetadataSendAlertMessage``  | Sends an alert message to the specified channel.                      |
    +---------------------------+---------------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | timeout                   | timeout                   | 3         | ``AutoModerationActionMetadataTimeout``           | Timeouts the user. Only applicable for `keyword` rules. Max 4 weeks.  |
    +---------------------------+---------------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | block_user_interaction    | block user interaction    | 4         | ``AutoModerationActionMetadataBase``              | Blocks the user from using text, voice and interactions.              |
    +---------------------------+---------------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    """
    __slots__ = ('metadata_type',)
    
    def __new__(cls, value, name = None, metadata_type = None):
        """
        Creates an auto moderation action type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the auto moderation action type.
        
        name : `None | str` = `None`, Optional
            The default name of the auto moderation action type.
        
        metadata_type : `None | type<AutoModerationActionMetadataBase>` = `None`, Optional
            The action type's respective metadata type.
        """
        if metadata_type is None:
            metadata_type = AutoModerationActionMetadataBase
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.metadata_type = metadata_type
        return self
    
    
    # predefined
    none = P(0, 'none', AutoModerationActionMetadataBase)
    block_message = P(1, 'block message', AutoModerationActionMetadataBlock)
    send_alert_message = P(2, 'send alert message', AutoModerationActionMetadataSendAlertMessage)
    timeout = P(3, 'timeout', AutoModerationActionMetadataTimeout)
    block_user_interaction = P(4, 'block user interaction', AutoModerationActionMetadataBase)



__all__ = ()

LIMIT_GLOBAL = 100
LIMIT_GUILD = 100
LENGTH_MAX = 4000

NAME_LENGTH_MIN = 1
NAME_LENGTH_MAX = 32
DESCRIPTION_LENGTH_MIN = 2
DESCRIPTION_LENGTH_MAX = 100
OPTIONS_MAX = 25
CHOICES_MAX = 25



__all__ = ('ApplicationCommandOption',)

from scarletio import RichAttributeErrorBaseType, copy_docs, export

from ...localization.helpers import get_localized_length
from ...localization.utils import hash_locale_dictionary

from ..application_command_option_metadata import ApplicationCommandOptionMetadataBase
from ..helpers import with_translation

from .fields import (
    parse_description, parse_description_localizations, parse_name, parse_name_localizations, parse_type,
    put_description, put_description_localizations, put_name, put_name_localizations, put_type,
    validate_description, validate_description_localizations, validate_name, validate_name_localizations, validate_type
)
from .helpers import _purge_defaults_and_maybe_raise
from .preinstanced import ApplicationCommandOptionType


@export
class ApplicationCommandOption(RichAttributeErrorBaseType):
    """
    An option of an ``ApplicationCommand``.
    
    Attributes
    ----------
    description : `None`, `str`
        The description of the application command option. It's length can be in range [1:100].
    
    description_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localized descriptions of the option.
    
    metadata : ``ApplicationCommandOptionMetadataBase``
        Metadata containing the type specific fields of the option.
    
    name : `str`
        The name of the application command option. It's length can be in range [1:32].
    
    name_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localized names of the option.
    
    type : ``ApplicationCommandOptionType``
        The option's type.
    """
    __slots__ = ('description', 'description_localizations', 'metadata', 'name', 'name_localizations', 'type')
    
    def __new__(
        cls,
        name,
        description,
        option_type,
        *,
        description_localizations = ...,
        name_localizations = ...,
        **keyword_parameters,
    ):
        """
        Creates a new application command option with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the command. It's length can be in range [1:32].
        
        description : `None`, `str`
            The command's description. It's length can be in range [2:100].
        
        option_type : `ApplicationCommandOptionType | int`
            The application command option's type.
        
        description_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`), Optional (Keyword only)
            Localized descriptions of the option.
        
        name_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`), Optional (Keyword only)
            Localized names of the option.
        
        **keyword_parameters : Keyword parameters
            Additional option type specific parameters.
        
        Other Parameters
        ----------------
        autocomplete : `bool`, Optional (Keyword only)
            Whether the option supports auto completion.
            
            Mutually exclusive with the `choices` parameter. Only applicable for string type parameters.
        
        channel_types : `None`, `iterable` of (``ChannelType``, `int`), Optional (Keyword only)
            The accepted channel types by the option.
            
            Only applicable if ``.type`` is set to `ApplicationCommandOptionType.channel`.
        
        choices : `None`, `iterable` of ``ApplicationCommandOptionChoice``, Optional (Keyword only)
            The choices of the command for string or integer types. It's length can be in range [0:25].
            
            Mutually exclusive with the `autocomplete` parameter.
        
        default : `bool` = `False`, Optional (Keyword only)
            Whether the option is the default one.
        
            
        max_length : `None`, `int`, Optional (Keyword only)
            The maximum input length allowed for this option.
            
            Only applicable for string options.
        
        max_value : `None`, `int`, `float`, Optional (Keyword only)
            The maximal value permitted for this option.
            
            Only applicable for integer as `int`, and for float options as `float`.
        
        min_length : `None`, `int`, Optional (Keyword only)
            The minimum input length allowed for this option.
            
            Only applicable for string options.
            
        min_value : `None`, `int`, `float`, Optional (Keyword only)
            The minimum value permitted for this option.
            
            Only applicable for integer as `int`, and for float options as `float`.
        
        options : `None`, `iterable` of ``ApplicationCommandOption``, Optional (Keyword only)
            The parameters or sub-commands of the command option. It's length can be in range [0:25].
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra parameters given.
        ValueError
            - If a parameter's value is incorrect.
            - if a parameter is not applicable for the given option.
        """
        # option_type
        option_type = validate_type(option_type)
        
        # description
        description = validate_description(description)
        
        # description_localizations
        if description_localizations is ...:
            description_localizations = None
        else:
            description_localizations = validate_description_localizations(description_localizations)
        
        # name
        name = validate_name(name)
        
        # name_localizations
        if name_localizations is ...:
            name_localizations = None
        else:
            name_localizations = validate_name_localizations(name_localizations)
        
        metadata = option_type.metadata_type.from_keyword_parameters(keyword_parameters)
        _purge_defaults_and_maybe_raise(keyword_parameters)
        
        # Construct
        self = object.__new__(cls)
        self.description = description
        self.description_localizations = description_localizations
        self.metadata = metadata
        self.name = name
        self.name_localizations = name_localizations
        self.type = option_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new application command option from the received data from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received application command option data.
        
        Returns
        -------
        self : `instance<type<cls>>`
            The created application command option.
        """
        option_type = parse_type(data)
        metadata = option_type.metadata_type.from_data(data)
        
        self = object.__new__(cls)
        self.description = parse_description(data)
        self.description_localizations = parse_description_localizations(data)
        self.metadata = metadata
        self.name = parse_name(data)
        self.name_localizations = parse_name_localizations(data)
        self.type = option_type
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the application command option to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = self.metadata.to_data(defaults = defaults)
        put_description(self.description, data, defaults)
        put_description_localizations(self.description_localizations, data, defaults)
        put_name(self.name, data, defaults)
        put_name_localizations(self.name_localizations, data, defaults)
        put_type(self.type, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the application command option's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields `.name`, `.description`, `.type`
        
        # name
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # description
        repr_parts.append(', description = ')
        repr_parts.append(repr(self.description))
        
        # type
        repr_parts.append(', type = ')
        option_type = self.type
        repr_parts.append(option_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(option_type.value))
        
        # Extra fields: `.autocomplete`, `.min_value`, `.max_value`, `.default`, `.required`, `.choices`, `.options`
        #    `.name_localizations`, `.description_localizations`, `.min_length`, `.max_length`
        
        # autocomplete
        if self.autocomplete:
            repr_parts.append(', autocomplete = True')
        
        # default
        if self.default:
            repr_parts.append(', default = True')
        
        # required
        if self.required:
            repr_parts.append(', required = True')
        
        if option_type is ApplicationCommandOptionType.string:
            # min_length
            min_length = self.min_length
            if (min_length != 0):
                repr_parts.append(', min_length = ')
                repr_parts.append(repr(min_length))
            
            # max_length
            max_length = self.max_length
            if (max_length != 0):
                repr_parts.append(', max_length = ')
                repr_parts.append(repr(max_length))
        
        
        if option_type is ApplicationCommandOptionType.integer or option_type is ApplicationCommandOptionType.float:
            # min_value
            min_value = self.min_value
            if (min_value is not None):
                repr_parts.append(', min_value = ')
                repr_parts.append(repr(min_value))
            
            # max_value
            max_value = self.max_value
            if (max_value is not None):
                repr_parts.append(', max_value = ')
                repr_parts.append(repr(max_value))
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            repr_parts.append(', channel_types = [')
            
            index = 0
            limit = len(channel_types)
            
            while True:
                channel_type = channel_types[index]
                index += 1
                repr_parts.append(channel_type.name)
                repr_parts.append(' ~ ')
                repr_parts.append(repr(channel_type.value))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # choices
        choices = self.choices
        if (choices is not None):
            repr_parts.append(', choices = [')
            
            index = 0
            limit = len(choices)
            
            while True:
                choice = choices[index]
                index += 1
                repr_parts.append(repr(choice))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        # options
        options = self.options
        if (options is not None):
            repr_parts.append(', options = [')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            repr_parts.append(', name_localizations = ')
            repr_parts.append(repr(name_localizations))
        
        # description_localizations
        description_localizations = self.description_localizations
        if (description_localizations is not None):
            repr_parts.append(', description_localizations = ')
            repr_parts.append(repr(description_localizations))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # description
        if self.description != other.description:
            return False
        
        # description_localizations
        if self.description_localizations != other.description_localizations:
            return False
        
        # metadata
        if self.metadata != other.metadata:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # name_localizations
        if self.name_localizations != other.name_localizations:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        return True
    
    
    def __len__(self):
        """Returns the application command option's length."""
        length = 0
        
        # choices
        for choice in self.iter_choices():
            length += len(choice)
        
        # description & description_localizations
        length += get_localized_length(self.description, self.description_localizations)
        
        # name & name_localizations
        length += get_localized_length(self.name, self.name_localizations)
        
        # options
        for option in self.iter_options():
            length += len(option)
        
        return length
    
    
    def __hash__(self):
        """Returns the hash value of the application command option."""
        hash_value = 0
        
        # autocomplete
        hash_value ^= self.autocomplete
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            hash_value ^= hash(channel_types)
        
        # choices
        choices = self.choices
        if (choices is not None):
            hash_value ^= len(choices) << 1
            
            for choice in choices:
                hash_value ^= hash(choice)
        
        # default
        hash_value ^= self.default << 5
        
        # description
        # Do not hash `.description` if equals to `.name`.
        description = self.description
        if (description != self.name):
            hash_value ^= hash(description)
        
        # description_localizations
        # Do not hash `.description_localizations` if equals to `.name_localizations`.
        description_localizations = self.description_localizations
        if (description_localizations is not None) and (description_localizations != self.name_localizations):
            hash_value ^= hash_locale_dictionary(description_localizations)
        
        # max_length
        hash_value ^= self.max_length << 9
        
        # max_value
        hash_value ^= hash(self.max_value)
        
        # min_length
        hash_value ^= self.min_length << 13
        
        # min_value
        hash_value ^= hash(self.min_value)
        
        # name
        hash_value ^= hash(self.name)
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            hash_value ^= hash_locale_dictionary(name_localizations)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options) << 1
            
            for option in options:
                hash_value ^= hash(option)
        
        # required
        hash_value ^= self.required << 17
        
        # type
        hash_value ^= self.type.value << 18
        
        return hash_value
    
    
    def with_translation(self, translation_table, replace = False):
        """
        Returns a new application command option with the given translation table applied.
        
        Parameters
        ----------
        translation_table : `None`, `dict` of ((``Locale``, `str`),
                (`None`, `dict` (`str`, (`None`, `str`)) items)) items
            Translation table to pull localization. from.
        replace : `bool` = `False`, Optional
            Whether actual translation should be replaced.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        if translation_table is None:
            return self
        
        new = self.copy()
        
        # choices
        choices = new.choices
        if (choices is not None):
            new.choices = (*(choice.with_translation(translation_table, replace) for choice in choices),)
        
        # description_localizations
        new.description_localizations = with_translation(
            new.description,
            new.description_localizations,
            translation_table,
            replace,
        )
        
        # name_localizations
        new.name_localizations = with_translation(
            new.name,
            new.name_localizations,
            translation_table,
            replace,
        )
        
        # options
        options = new.options
        if (options is not None):
            new.options = (*(option.with_translation(translation_table, replace) for option in options),)
        
        return new
    
    
    def copy(self):
        """
        Copies the application command option.
        
        Returns
        -------
        new : `instance<type<cls>>`
        """
        new = object.__new__(type(self))
        
        # description
        new.description = self.description
        
        # description_localizations
        description_localizations = self.description_localizations
        if (description_localizations is not None):
            description_localizations = description_localizations.copy()
        new.description_localizations = description_localizations
        
        # metadata
        new.metadata = self.metadata.copy()
        
        # name
        new.name = self.name
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            name_localizations = name_localizations.copy()
        new.name_localizations = name_localizations
        
        # type
        new.type = self.type
        
        return new
    

    def copy_with(
        self,
        *,
        description = ...,
        description_localizations = ...,
        name = ...,
        name_localizations = ...,
        option_type = ...,
        **keyword_parameters,
    ):
        """
        Copies the application command option wit the given fields.
        
        Parameters
        ----------
        description : `None`, `str`, Optional (Keyword only)
            The command's description. It's length can be in range [2:100].
        
        description_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`), Optional (Keyword only)
            Localized descriptions of the option.
            
        name : `str`, Optional (Keyword only)
            The name of the command. It's length can be in range [1:32].
        
        name_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`), Optional (Keyword only)
            Localized names of the option.
        
        option_type : `int`, ``ApplicationCommandOptionType``, Optional (Keyword only)
            The application command option's type.
        
        **keyword_parameters : Keyword parameters
            Additional option type specific parameters.
        
        Other Parameters
        ----------------
        autocomplete : `bool`, Optional (Keyword only)
            Whether the option supports auto completion.
            
            Mutually exclusive with the `choices` parameter. Only applicable for string type parameters.
        
        channel_types : `None`, `iterable` of (``ChannelType``, `int`), Optional (Keyword only)
            The accepted channel types by the option.
            
            Only applicable if ``.type`` is set to `ApplicationCommandOptionType.channel`.
        
        choices : `None`, `iterable` of ``ApplicationCommandOptionChoice``, Optional (Keyword only)
            The choices of the command for string or integer types. It's length can be in range [0:25].
            
            Mutually exclusive with the `autocomplete` parameter.
        
        default : `bool` = `False`, Optional (Keyword only)
            Whether the option is the default one.
        
        max_length : `None`, `int`, Optional (Keyword only)
            The maximum input length allowed for this option.
            
            Only applicable for string options.
        
        max_value : `None`, `int`, `float`, Optional (Keyword only)
            The maximal value permitted for this option.
            
            Only applicable for integer as `int`, and for float options as `float`.
        
        min_length : `None`, `int`, Optional (Keyword only)
            The minimum input length allowed for this option.
            
            Only applicable for string options.
            
        min_value : `None`, `int`, `float`, Optional (Keyword only)
            The minimum value permitted for this option.
            
            Only applicable for integer as `int`, and for float options as `float`.
        
        options : `None`, `iterable` of ``ApplicationCommandOption``, Optional (Keyword only)
            The parameters or sub-commands of the command option. It's length can be in range [0:25].
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Returns
        -------
        new : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra parameters given.
        ValueError
            - If a parameter's value is incorrect.
            - if a parameter is not applicable for the given option.
        """
        # option_type
        if (option_type is ...):
            option_type = self.type
        else:
            option_type = validate_type(option_type)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # description_localizations
        if description_localizations is ...:
            description_localizations = self.description_localizations
            if (description_localizations is not None):
                description_localizations = description_localizations.copy()
        else:
            description_localizations = validate_description_localizations(description_localizations)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # name_localizations
        if name_localizations is ...:
            name_localizations = self.name_localizations
            if (name_localizations is not None):
                name_localizations = name_localizations.copy()
        else:
            name_localizations = validate_name_localizations(name_localizations)
        
        # metadata
        metadata = self.metadata
        metadata_type = option_type.metadata_type
        if metadata_type is type(metadata):
            metadata = metadata.copy_with_keyword_parameters(keyword_parameters)
        else:
            metadata = metadata_type.from_keyword_parameters(keyword_parameters)
        _purge_defaults_and_maybe_raise(keyword_parameters)
        
        
        # Construct
        new = object.__new__(type(self))
        new.description = description
        new.description_localizations = description_localizations
        new.metadata = metadata
        new.name = name
        new.name_localizations = name_localizations
        new.type = option_type
        return new
    
    # Field proxies
    
    # autocomplete
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.autocomplete)
    def autocomplete(self):
        return self.metadata.autocomplete
    
    @autocomplete.setter
    def autocomplete(self, autocomplete):
        self.metadata.autocomplete = autocomplete
    
    # channel_types
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.channel_types)
    def channel_types(self):
        return self.metadata.channel_types
    
    @channel_types.setter
    def channel_types(self, channel_types):
        self.metadata.channel_types = channel_types
    
    # choices
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.choices)
    def choices(self):
        return self.metadata.choices
    
    @choices.setter
    def choices(self, choices):
        self.metadata.choices = choices
    
    # default
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.default)
    def default(self):
        return self.metadata.default
    
    @default.setter
    def default(self, default):
        self.metadata.default = default
    
    # max_length
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.max_length)
    def max_length(self):
        return self.metadata.max_length
    
    @max_length.setter
    def max_length(self, max_length):
        self.metadata.max_length = max_length
    
    # max_value
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.max_value)
    def max_value(self):
        return self.metadata.max_value
    
    @max_value.setter
    def max_value(self, max_value):
        self.metadata.max_value = max_value
    
    # min_length
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.min_length)
    def min_length(self):
        return self.metadata.min_length
    
    @min_length.setter
    def min_length(self, min_length):
        self.metadata.min_length = min_length
    
    # min_value
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.min_value)
    def min_value(self):
        return self.metadata.min_value
    
    @min_value.setter
    def min_value(self, min_value):
        self.metadata.min_value = min_value
    
    # options
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.options)
    def options(self):
        return self.metadata.options
    
    @options.setter
    def options(self, options):
        self.metadata.options = options
    
    # required
    @property
    @copy_docs(ApplicationCommandOptionMetadataBase.required)
    def required(self):
        return self.metadata.required
    
    @required.setter
    def required(self, required):
        self.metadata.required = required
    
    # Iterators
    
    def iter_choices(self):
        """
        Iterates over the choices of the application command option.
        
        This method is an iterable generator.
        
        Yields
        ------
        choice : ``ApplicationCommandOptionChoice``
        """
        choices = self.choices
        if (choices is not None):
            yield from choices
    
    
    def iter_options(self):
        """
        Iterates over the options of the application command option.
        
        This method is an iterable generator.
        
        Yields
        ------
        option : ``ApplicationCommandOption``
        """
        options = self.options
        if (options is not None):
            yield from options
    



__all__ = ()


def build_edit_payload(old_entity, new_entity, field_converters, keyword_parameters):
    """
    Edit payload builder.
    
    Parameters
    ----------
    old_entity : `None`, `object with to_data(defaults)`
        Original entity.
    
    new_entity : `None` `object with to_data(defaults)`
        New entity to serialize to.
    
    field_converters : `dict` of (`str`, `tuple` (`callable`, `callable`)) items
        Fields converters to use.
    
    keyword_parameters : `dict` of (`str`, `object`) items
        Keyword parameters passed to the original function.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
        The built payload.
    
    Raises
    ------
    TypeError
        - Extra or unused parameter.
    """
    if (new_entity is None):
        data = {}
    else:
        data = new_entity.to_data(defaults = True)
        
        if (old_entity is not None):
            for old_field_name, old_field_value in old_entity.to_data(defaults = True).items():
                try:
                    new_field_value = data[old_field_name]
                except KeyError:
                    pass
                else:
                    if (old_field_value == new_field_value):
                        del data[old_field_name]
    
    add_payload_fields_from_keyword_parameters(field_converters, keyword_parameters, data, True)
    return data


def build_create_payload(entity, field_converters, keyword_parameters):
    """
    Create payload builder.
    
    Parameters
    ----------
    entity : `None`, `object with to_data(defaults)`
        An entity which can be serialized.
    
    field_converters : `dict` of (`str`, `tuple` (`callable`, `callable`)) items
        Fields converters to use.
    
    keyword_parameters : `dict` of (`str`, `object`) items
        Keyword parameters passed to the original function.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
        The built payload.
    
    Raises
    ------
    TypeError
        - Extra or unused parameter.
    """
    if (entity is None):
        data = {}
    else:
        data = entity.to_data(defaults = False)
    
    if keyword_parameters:
        add_payload_fields_from_keyword_parameters(field_converters, keyword_parameters, data, False)
    
    return data


def add_payload_fields_from_keyword_parameters(
    field_converters, keyword_parameters, data, defaults, *, raise_unused = True
):
    """
    Generic payload builder used by ``build_edit_payload`` and by ``build_create_payload``.
    
    Parameters
    ----------
    field_converters : `dict` of (`str`, `tuple` (`callable`, `callable`)) items
        Fields converters to use.
    
    keyword_parameters : `dict` of (`str`, `object`) items
        Keyword parameters passed to the original function.
    
    data : `dict` of (`str`, `object`) items
        The built payload.
    
    defaults : `bool`
        Whether default values should be included.
    
    raise_unused : `bool` = `True`, Optional (Keyword only)
        Whether exception should be raised when there are unused keyword parameters.
    
    Raises
    ------
    TypeError
        - Extra or unused parameter.
    
    Returns
    -------
    unused : `None | dict<str, object>`
    """
    unused = None
    
    while keyword_parameters:
        field_name, field_value = keyword_parameters.popitem()
        
        try:
            validator, putter = field_converters[field_name]
        except KeyError:
            if unused is None:
                unused = {}
            
            unused[field_name] = field_value
            continue
        
        putter(validator(field_value), data, defaults)
    
    if raise_unused and (unused is not None):
        raise TypeError(
            f'Unused or extra parameters: {unused!r}.'
        )
    return unused
